import tritonclient.http as httpclient
from tritonclient.utils import np_to_triton_dtype
from abottle.base_model import BaseModel


class TritonModel(BaseModel):
    class Config:
        name = ""
        version = ""
        triton_url = "trition.trition-system"

    def __init__(self):
        self.triton_client = httpclient.InferenceServerClient(
            url=self.Config.triton_url, insecure=True
        )

    def infer(self, X={}, Y=[]):
        inputs = []
        for x_name, x in X.items():
            infer_input = httpclient.InferInput(
                x_name, x.shape, np_to_triton_dtype(x.dtype)
            )
            infer_input.set_data_from_numpy(x)
            inputs.append(infer_input)

        outputs = []
        for y in Y:
            outputs.append(httpclient.InferRequestedOutput(y))
        results = self.triton_client.infer(
            self.Config.name,
            inputs,
            model_version=str(self.Config.version),
            outputs=outputs,
        )

        outputs = {}
        for y in Y:
            outputs[y] = results.as_numpy(y)
        return outputs



import json
import typing
from json import JSONEncoder
import numpy


class NumpyEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, numpy.ndarray):
            return obj.tolist()
        return JSONEncoder.default(self, obj)


from fastapi import FastAPI, Request, Body
from fastapi.responses import Response
from pydantic import schema_of
import importlib
import uvicorn
from typing import Any

app = FastAPI()

USERMODEL = None


@app.post("/predict")
async def predict(request: Request, X: typing.Any = Body(...)):
    user_request = await request.json()
    # FIXME: should comprese resposne content
    user_response = USERMODEL.predict(**user_request)
    response_content = json.dumps(user_response, cls=NumpyEncoder)
    # log here
    return Response(content=response_content, media_type="application/json")


@app.post("/metrics")
async def metrics():
    raise NotImplementedError()


@app.get("/healthz")
async def healthz():
    if hasattr(USERMODEL, "status"):
        return USERMODEL.status()
    return True


@app.get("/readyz")
async def readyz():
    if hasattr(USERMODEL, "status"):
        return USERMODEL.status()
    return True


@app.get("/meta")
async def meta():
    if hasattr(USERMODEL, "meta"):
        user_meta = USERMODEL.meta()
        if user_meta:
            return user_meta
    user_meta = USERMODEL.predict.__annotations__
    user_meta_dict = {}
    for key in user_meta:
        user_meta_dict[key] = schema_of(user_meta[key], title=key)
    return user_meta_dict


from pathlib import Path
import importlib
import argparse
import os
import sys


def locate(usermodel_name):
    package_dir = Path(__name__).resolve().parent
    sys.path.append(str(package_dir))
    usermodel_names = usermodel_name.split(".")
    module = importlib.import_module(usermodel_names[0])
    my_class = None
    for attr in usermodel_names[1:]:
        my_class = getattr(module, attr)
    return my_class


from yaml import load as yaml_load

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper


class Struct(object):
    def __init__(self, data):
        for name, value in data.items():
            setattr(self, name, self._wrap(value))

    def _wrap(self, value):
        if isinstance(value, (tuple, list, set, frozenset)):
            return type(value)([self._wrap(v) for v in value])
        else:
            return Struct(value) if isinstance(value, dict) else value


def load(usermodel_name, wrapper, config=None, *args, **kwargs):

    config_data = None
    config_obj = None

    if config:
        try:
            with open(config, "r") as f:
                config_data = yaml_load(f, Loader=Loader)
        except FileNotFoundError:
            config_data = yaml_load(config, Loader=Loader)

        config_obj = Struct(config_data)

    usermodel = locate(usermodel_name)()

    if config_obj and not hasattr(usermodel, "Config"):
        usermodel.Config = config_obj
    elif not hasattr(usermodel, "Config"):
        print("your user model contain Config filed, will ignore --config argument")

    wrapper_cls = locate(wrapper)
    config_filed_name = wrapper.split(".")[-1]

    if hasattr(usermodel.Config, config_filed_name):
        usermodel = wrapper_cls.load_from(usermodel)

    else:
        raise Exception(
            f"no config field {config_filed_name} found in your model, as the wrapper {wrapper} you passed, {config_filed_name} will be your config filed name, and it must exsits in your config"
        )
    return usermodel


def start_server(host, port):
    uvicorn.run(app, host=host, port=port)


def run_tester(**kwargs):
    signatures = USERMODEL.evaluate.__func__.__annotations__
    pass_kwargs = {}
    for (
        signature_name,
        signature_type,
    ) in signatures.items():
        value = kwargs.get(signature_name, None)
        if signature_name == "return":
            continue
        if (
            not value
        ):  # FIXME: <signature_name> may means a kwarg with default name, but in this code, we assume it as a error
            raise Exception(f"you should set {signature_name}")
        pass_kwargs[signature_name] = signature_type(value)
    return USERMODEL.evaluate(**pass_kwargs)


def main():
    parser = argparse.ArgumentParser(description="Warp you python object with a bottle")
    parser.add_argument("usermodel_name", help="your python object moudle")
    parser.add_argument(
        "--wrapper",
        help="which model wrapper you want to use? abottle.TritonModel? abottle.ONNXModel, abottle.TensorRTModel?, abottle.PytrochModel? or any wrapper class that implemented abottle.BaseModel!",
        default="abottle.TritonModel",
    )
    parser.add_argument("--as", dest="as_", help="server? tester?", default="server")
    parser.add_argument(
        "--config",
        required=False,
        dest="config",
        help="config yaml file path or content in string",
    )
    parser.add_argument("--host", default="0.0.0.0")
    parser.add_argument("--port", default=8081)
    # FIXME: shuold set file path and batch size as fix argument?

    args, unknowns = parser.parse_known_args()
    global USERMODEL
    USERMODEL = load(**vars(args))

    if args.as_ == "server":
        start_server(args.host, args.port)
    if args.as_ == "tester":
        tester_args = {}
        for unknown in unknowns:
            unknown.replace("-", "")
            k, v = unknown.split("=")
            tester_args[k] = v
        run_tester(**tester_args)


if __name__ == "__main__":
    main()



try:
    from abottle.trt_model import TensorRTModel
except ImportError:
    print("TensorRTModel import failed, ignoreing...")

try:
    from abottle.onnx_model import ONNXModel
except ImportError:
    print("ONNXModel import failed, ignoreing...")

try:
    from abottle.triton_model import TritonModel
except ImportError:
    print("TritonModel import failed, ignoreing...")

try:
    from abottle.pytorch_model import PytorchModel
except ImportError:
    print("PytorchModel import failed, ignoreing...")

from abottle.base_model import BaseModel



import onnxruntime as ort
from abottle.base_model import BaseModel


class ONNXModel(BaseModel):
    class Config:
        ort_file = ""

    def __init__(self):
        self.session = ort.InferenceSession(
            self.Config.ort_file,
            providers=["CUDAExecutionProvider"],
        )

    def infer(self, X={}, Y=[]):
        return self.session.run(None, X)



import os
import tensorrt as trt

TRT_LOGGER = trt.Logger()


def get_engine(engine_file_path):
    print("Reading engine from file {}".format(engine_file_path))
    with open(engine_file_path, "rb") as f, trt.Runtime(TRT_LOGGER) as runtime:
        return runtime.deserialize_cuda_engine(f.read())


class HostDeviceMem(object):
    def __init__(self, host_mem, device_mem):
        self.host = host_mem
        self.device = device_mem

    def __str__(self):
        return "Host:\n" + str(self.host) + "\nDevice:\n" + str(self.device)

    def __repr__(self):
        return self.__str__()


import pycuda.autoinit
import pycuda.driver as cuda


class allocator:
    def __init__(self, engine):
        self.engine = engine

    def __enter__(self):
        self.inputs = []
        self.outputs = []
        self.bindings = []
        self.stream = cuda.Stream()
        for binding in self.engine:
            size = (
                trt.volume(self.engine.get_binding_shape(binding))
                * self.engine.max_batch_size
            )
            dtype = trt.nptype(self.engine.get_binding_dtype(binding))
            # Allocate host and device buffers
            host_mem = cuda.pagelocked_empty(size, dtype)
            device_mem = cuda.mem_alloc(host_mem.nbytes)
            # Append the device buffer to device bindings.
            self.bindings.append(int(device_mem))
            # Append to the appropriate list.
            if self.engine.binding_is_input(binding):
                self.inputs.append(HostDeviceMem(host_mem, device_mem))
            else:
                self.outputs.append(HostDeviceMem(host_mem, device_mem))

        return self.inputs, self.outputs, self.bindings, self.stream

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self.inputs, self.outputs, self.bindings, self.stream


def do_inference(context, bindings, inputs, outputs, stream):
    # Transfer input data to the GPU.
    [cuda.memcpy_htod_async(inp.device, inp.host, stream) for inp in inputs]
    # Run inference.
    context.execute_async_v2(bindings=bindings, stream_handle=stream.handle)
    # Transfer predictions back from the GPU.
    [cuda.memcpy_dtoh_async(out.host, out.device, stream) for out in outputs]
    # Synchronize the stream
    stream.synchronize()
    # Return only the host outputs.
    return [out.host for out in outputs]


from abottle.base_model import BaseModel


class TensorRTModel(BaseModel):
    class Config:
        trt_file = ""

    def __init__(self):
        self.engine = get_engine(self.Config.trt_file)
        self.context = self.engine.create_execution_context()

    def __del__(self):
        del self.context, self.engine

    def infer(self, X={}, Y=[]):

        with allocator(self.engine) as (inputs, outputs, bindings, stream):
            for i, (k, v) in enumerate(X.items()):
                inputs[i].host = v

            return do_inference(
                self.context,
                bindings=bindings,
                inputs=inputs,
                outputs=outputs,
                stream=stream,
            )



class BaseModel:
    class Config:
        """you can set any field you like in your config class"""

        pass

    def infer(self, X={}, Y=[]):
        """infer function you can do anything here"""
        raise NotImplementedError()

    def evaluate(self, *args, **kwargs):
        raise NotImplementedError()

    @classmethod
    def load_from(cls, usermodel):
        """you shuold set your wrapper from usermodel, and set Config like"""
        cls.Config = getattr(usermodel.Config, cls.__name__)
        usermodel.model = cls()
        return usermodel
